#############################################
# calcul de carte de disparites, version 0.5
# pour blender 2.34
#############################################

import Blender
from Blender import *
from math import *
from Blender.Mathutils import *
from Blender.Scene import Render
import math
import gc
import os
import time
import sys

pi = 3.1415926565897932
eps = 1e-6

## ####################################
######  arguments d'entree     #######
######################################
###outpath = "D:/LABO/BLENDER/PLANTE/"
outpath = "/home/bocquill/BLENDER/PLANTE/"
baseline = 0.5
name_left = "image_g.jpg"
name_right = "image_d.jpg"
cam_pos = 0
render_images = 0
calc_disp = 1
xres = 400
yres = 300
################################################


###################################################
#            fonctions annexes
###################################################

# fonction qui multiplie 2 "matrices"
def SMatMultMat(M,N):
	P = []
	for i in range(len(M)):
		L = []
		for j in range(len(N[0])):
			s = 0
			for k in range(len(M[0])):
				s = s + M[i][k]*N[k][j]
			L.append(s)
		P.append(L)
	del L
	return P

# fonction qui affiche une "matrice"
def SPrintMat(M):
	s = ""
	for i in range(len(M)):
		print M[i]
	return s

# fonction qui multiplie une "matrice" et un "vecteur"
def SMatMultVec(M,V):
	W = []
	for i in range(len(M)):
		s = 0
		for k in range(len(M[0])):
			s = s + M[i][k]*V[k]
		W.append(s)
	return W		

# fonction qui transforme une matrice de transformation de sorte que p' = Mp
def STransformationMat(m):
	return [[m[0][0],m[1][0],m[2][0],m[3][0]], [m[0][1],m[1][1],m[2][1],m[3][1]], [m[0][2],m[1][2],m[2][2],m[3][2]], [0,0,0,1]]

# fonction qui normalise un vecteur homogene
def SNormalise_homogene(v):
	if len(v) == 3:
		v[0] = v[0] / v[2]
		v[1] = v[1] / v[2]
		v[2] = 1
	else:
		v[0] = v[0] / v[3]
		v[1] = v[1] / v[3]
		v[2] = v[2] / v[3]
		v[3] = 1

# fonction qui calcule le produit vectoriel de 2 "vecteurs"
def SCrossVecs(v,w):
	z = []
	z.append(v[1]*w[2]-v[2]*w[1])
	z.append(v[2]*w[0]-v[0]*w[2])
	z.append(v[0]*w[1]-v[1]*w[0])
	return z

# fonction qui transpose 1 matrice
def STransMat(M):
	P = []
	for i in range(len(M[0])):
		L = []
		for j in range(len(M)):
			L.append(M[j][i])
		P.append(L)
	del L
	return P

# fonction qui normalise un vecteur
def SNormalise(v):
	if len(v) == 3:
		l = math.sqrt(v[0] * v[0] + v[1] * v[1] + v[2] * v[2])
		if l <= eps:
			print "error in normalize"
			v[0] = 0
			v[1] = 1
			v[2] = 0
		else:
			v[0] = v[0] / l
			v[1] = v[1] / l
			v[2] = v[2] / l
	else:
		l = math.sqrt(v[0] * v[0] + v[1] * v[1] + v[2] * v[2] + v[3]*v[3])
		if l <= eps:
			print "error in normalize"
			v[0] = 0
			v[1] = 1
			v[2] = 0
			v[3] = 0
		else:
			v[0] = v[0] / l
			v[1] = v[1] / l
			v[2] = v[2] / l
			v[3] = v[3] / l

#fonction qui calcule le determinant d'une matrice
def SDetMat(m):
	return (m[0][0] * (m[1][1] * m[2][2] - m[2][1] * m[1][2]) - m[1][0] * (m[0][1] * m[2][2] - m[2][1] * m[0][2]) + m[2][0] * (m[0][1] * m[1][2] - m[1][1] * m[0][2]))

# fonction qui calcule l'inverse d'une matrice 3x3 ou 4x4
def SInvMat(m):
	det = (m[0][0] * (m[1][1] * m[2][2] - m[2][1] * m[1][2]) - m[1][0] * (m[0][1] * m[2][2] - m[2][1] * m[0][2]) + m[2][0] * (m[0][1] * m[1][2] - m[1][1] * m[0][2]))
	if det == 0.0: 
		return None
	det = 1.0 / det
	if len(m) == 4:
		r = [ [det * (m[1][1] * m[2][2] - m[2][1] * m[1][2]), - det * (m[0][1] * m[2][2] - m[2][1] * m[0][2]),det * (m[0][1] * m[1][2] - m[1][1] * m[0][2]),0.0,], [- det * (m[1][0] * m[2][2] - m[2][0] * m[1][2]),det * (m[0][0] * m[2][2] - m[2][0] * m[0][2]),- det * (m[0][0] * m[1][2] - m[1][0] * m[0][2]),0.0], [det * (m[1][0] * m[2][1] - m[2][0] * m[1][1]),- det * (m[0][0] * m[2][1] - m[2][0] * m[0][1]),det * (m[0][0] * m[1][1] - m[1][0] * m[0][1]),0.0,] ]
		r.append([-(m[3][0] * r[0][0] + m[3][1] * r[1][0] + m[3][2] * r[2][0]),-(m[3][0] * r[0][1] + m[3][1] * r[1][1] + m[3][2] * r[2][1]),-(m[3][0] * r[0][2] + m[3][1] * r[1][2] + m[3][2] * r[2][2]),1.0,])
	else:
		r = [ [det * (m[1][1] * m[2][2] - m[2][1] * m[1][2]),- det * (m[0][1] * m[2][2] - m[2][1] * m[0][2]),det * (m[0][1] * m[1][2] - m[1][1] * m[0][2]),], [- det * (m[1][0] * m[2][2] - m[2][0] * m[1][2]),det * (m[0][0] * m[2][2] - m[2][0] * m[0][2]),- det * (m[0][0] * m[1][2] - m[1][0] * m[0][2]),], [det * (m[1][0] * m[2][1] - m[2][0] * m[1][1]),- det * (m[0][0] * m[2][1] - m[2][0] * m[0][1]),det * (m[0][0] * m[1][1] - m[1][0] * m[0][1]),] ]
	return r

# fonction qui calcule le produit scalaire de 2 "vecteurs"
def SDotVecs(v,w):
	return v[0]*w[0] + v[1]*w[1] + v[2]*w[2]

# fonction qui cree une matrice de rotation selon l'axe x
def SMatrix_rotate_x(angle):
	cos = math.cos(angle)
	sin = math.sin(angle)
	return [
	[1.0,  0.0, 0.0, 0.0],
	[0.0,  cos, sin, 0.0],
	[0.0, -sin, cos, 0.0],
	[0.0,  0.0, 0.0, 1.0],
	]

# fonction qui cree une matrice de rotation selon l'axe y
def SMatrix_rotate_y(angle):
	cos = math.cos(angle)
	sin = math.sin(angle)
	return [
	[cos, 0.0, -sin, 0.0],
	[0.0, 1.0,  0.0, 0.0],
	[sin, 0.0,  cos, 0.0],
	[0.0, 0.0,  0.0, 1.0],
	]

# fonction qui cree une matrice de rotation selon l'axe z
def SMatrix_rotate_z(angle):
	cos = math.cos(angle)
	sin = math.sin(angle)
	return [
 	[ cos, sin, 0.0, 0.0],
	[-sin, cos, 0.0, 0.0],
	[ 0.0, 0.0, 1.0, 0.0],
	[ 0.0, 0.0, 0.0, 1.0],
	]

# fonction qui retourne une matrice scale
def SMatrix_scale(fx, fy, fz):
	return [
	[ fx, 0.0, 0.0, 0.0],
	[0.0,  fy, 0.0, 0.0],
	[0.0, 0.0,  fz, 0.0],
	[0.0, 0.0, 0.0, 1.0],
	]
		
# fonction qui retourne la matrice d'un objet obj
def SGetMatrix(obj):
	rot = obj.rot
	mx = SMatrix_rotate_x(rot[0])
	my = SMatrix_rotate_y(rot[1])
	mz = SMatrix_rotate_z(rot[2])
	m = SMatMultMat(my,mz)
	m = SMatMultMat(mx,m)
	size = obj.size
	ms = SMatrix_scale(size[0],size[1],size[2])
	m = SMatMultMat(ms,m)
	loc = obj.loc
	m[3][0] = loc[0]
	m[3][1] = loc[1]
	m[3][2] = loc[2]
	del rot,loc,mx,my,mz,size,ms
	return m

# fonction qui recupere tous les meshs
def getAllMats(Rt):
	N = 0
	D = 0
	Me = 0
	objs = Blender.Object.Get()
	tmats = []
	for ob in objs:
		if ob.getType() == 'Mesh':
			Me = Me + 1
			m = SGetMatrix(ob)
			m = STransformationMat(m)
			Rtm = SMatMultMat(Rt,m)
			for face in ob.getData().faces:
				if len(face.v) > 4:
					print "un quad !"
				# equation du plan de la face
				s1 = face.v[0]
				s1 = [s1.co[0], s1.co[1], s1.co[2], 1]
				s1 = SMatMultVec(Rtm, s1)
				s2 = face.v[1]
				s2 = [s2.co[0], s2.co[1], s2.co[2], 1]
				s2 = SMatMultVec(Rtm, s2)
				s3 = face.v[2]
				s3 = [s3.co[0], s3.co[1], s3.co[2], 1]
				s3 = SMatMultVec(Rtm, s3)

				M = []
				M.append([s1[0], s2[0], s3[0]])
				M.append([s1[1], s2[1], s3[1]])
				M.append([s1[2], s2[2], s3[2]])
				M.append([1,1,1])
				
				m234 = [M[1],M[2],M[3]]
				d234 = SDetMat(m234)
				m134 = [M[0],M[2],M[3]]
				d134 = SDetMat(m134)
				m124 = [M[0],M[1],M[3]]
				d124 = SDetMat(m124)
				m123 = [M[0],M[1],M[2]]
				d123 = SDetMat(m123)
				PI = [d234,-d134,d124,-d123]
				# test du cas degenere
				if abs(d123) < 0.0001:
					D = D + 1
					continue						
				
				N = N + 1

				SNormalise_homogene(PI)
				del M,m234,d234,m134,d134,m124,d124,m123,d123
				del s1,s2,s3

				cotes = []
				# calcul des coordonnees de Plucker pour les cotes de la face
				for i in range(len(face.v)):
					u = face.v[i]
					if i != (len(face.v)-1):
						v = face.v[i+1]
					else:
						v = face.v[0]
					# transformation des vertex dans le repere camera
					u = [u[0], u[1], u[2], 1]
					u = SMatMultVec(Rtm,u)
					SNormalise_homogene(u)
					u = [u[0], u[1], u[2]]

					v = [v[0], v[1], v[2], 1]
					v = SMatMultVec(Rtm,v)
					SNormalise_homogene(v)
					v = [v[0], v[1], v[2]]

					c = []
					c.append([v[0]-u[0],v[1]-u[1],v[2]-u[2]])
					c.append(SCrossVecs(v,u))
					cotes.append(c)

				tmats.append([PI,cotes])
	print Me, " objets dans la scene"
	print N, " faces dans la scene"
	print D, " faces degenerees"
	return tmats

# fonction qui determine si deux segments s'intersectent
def Segment_intersect(ptA1, ptA2, ptB1, ptB2):
	lA = SCrossVecs(ptA1, ptA2)
	lB = SCrossVecs(ptB1, ptB2)
	x = SCrossVecs(lA, lB)
	if abs(x[2]) < eps:
		return 0
	SNormalise_homogene(x)
	if ptA1[0] <= ptA2[0]:
		xA1 = ptA1[0] - eps
		xA2 = ptA2[0] + eps
	else:
		xA1 = ptA2[0] - eps
		xA2 = ptA1[0] + eps
	if ptB1[0] <= ptB2[0]:
		xB1 = ptB1[0] - eps
		xB2 = ptB2[0] + eps
	else:
		xB1 = ptB2[0] - eps
		xB2 = ptB1[0] + eps
	if ptA1[1] <= ptA2[1]:
		yA1 = ptA1[1] - eps
		yA2 = ptA2[1] + eps
	else:
		yA1 = ptA2[1] - eps
		yA2 = ptA1[1] + eps
	if ptB1[1] <= ptB2[1]:
		yB1 = ptB1[1] - eps
		yB2 = ptB2[1] + eps
	else:
		yB1 = ptB2[1] - eps
		yB2 = ptB1[1] + eps
	if x[0] >= xA1 and x[0] <= xA2 and x[0] >= xB1 and x[0] <= xB2 and x[1] >= yA1 and x[1] <= yA2 and x[1] >= yB1 and x[1] <= yB2:
		return 1
	else:
		return 0

# fonction qui determine si un point est dans un segment
def Pt_in_segment(pt, ptA1, ptA2):
	# le point doit appartenir a la droite
	l = SCrossVecs(ptA1, ptA2)
	s = SDotVecs(l, pt)
	if abs(s) > eps:
		return 0
	if ptA1[0] <= ptA2[0]:
		x1 = ptA1[0] - eps
		x2 = ptA2[0] + eps
	else:
		x1 = ptA2[0] - eps
		x2 = ptA1[0] + eps
	if ptA1[1] <= ptA2[1]:
		y1 = ptA1[1] - eps
		y2 = ptA2[1] + eps
	else:
		y1 = ptA2[1] - eps
		y2 = ptA1[1] + eps
	if pt[0] >= x1 and pt[0] <= x2 and pt[1] >= y1 and pt[1] <= y2:
		return 1
	else:
		return 0

# fonction qui determine si un point est dans une face d'un mesh (= un polygone) (cotes et sommets sont consideres comme interieur du polygone)
def Pt_in_polygon(som, pt):
	xmax = som[0][0]
	for s in som:
		if s[0] > xmax:
			xmax = s[0]
	pt2 = [xmax+10,pt[1],1]
	N = 0
	for i in range(len(som)):
		if i != (len(som)-1):
			ptB1 = [som[i][0], som[i][1], 1]
			ptB2 = [som[i+1][0], som[i+1][1], 1]
		else:
			ptB1 = [som[i][0], som[i][1], 1]
			ptB2 = [som[0][0], som[0][1], 1]
		r = Pt_in_segment(pt, ptB1, ptB2)	
		if r == 1:
			return 1 
		r = Segment_intersect(pt, pt2, ptB1, ptB2)		
		N = N + r
	if fmod(N,2) != 0:
		return 1
	else:	
		return 0

# fonction qui donne le signe de X
def sign(X):
	if abs(X) == X:
		return 1
	else:
		return -1

# fonction qui calcule si un rayon intersecte une face
def RayFaceIntersect(cotes, rayon):
	s = 0
	for cote in cotes:
		w = SDotVecs(rayon[0],cote[1]) + SDotVecs(cote[0],rayon[1])
		#print w
		if s != 0:
			if sign(w) != s:
				return 0
		else:
			s = sign(w)
	return 1
			

###################################################
#            programme principal
###################################################
print "***** c''est parti "
# on recupere les objets scene et camera
curScene = Scene.getCurrent()
cam = curScene.getCurrentCamera()
cam_data = cam.getData()

# placement de la camera
if cam_pos == 1:
	cam.setLocation(xc, yc, zc)
	cam.RotX = rotxc
	cam.RotY = rotyc
	cam.RotZ = rotzc

cam_size = cam.getSize()
cam.setSize(1,1,1)

context = curScene.getRenderingContext()
context.setRenderPath(outpath)
context.imageSizeX(xres)
context.imageSizeY(yres)

# rendu image gauche
if render_images == 1:
### reglages avances du rendu ###
	"""
	context.setRenderer(Render.YAFRAY)
	context.yafrayGamma(1.3)
	context.setYafrayGIMethod(Render.SKYDOME)
	context.yafrayGIPower(7)
	"""
	context.startFrame(1)
	context.endFrame(1)
	context.renderAnim()
	try:
		os.rename(outpath+"0001.jpg",outpath+name_left)
	except OSError:
		os.remove(outpath+name_left)
		os.rename(outpath+"0001.jpg",outpath+name_left)	

	# on deplace la camera de baseline selon Xcamera pour la seconde image
	if cam_pos == 1:
		cam.LocX = cam.LocX + baseline
	else:
		v = Vector([baseline,0,0,1])
		m = cam.getInverseMatrix()
		m = STransformationMat(m)
		m = STransMat(m)
		vw = SMatMultVec(m,v)
		SNormalise_homogene(vw)
		vw = [vw[0], vw[1], vw[2]]
		SNormalise(vw)
		vw = [vw[0]*baseline,vw[1]*baseline,vw[2]*baseline]

		loc = cam.getLocation()
		cam.setLocation(loc[0]+vw[0],loc[1]+vw[1],loc[2]+vw[2])
	curScene.update()

	# rendu image droite
	context.startFrame(1)
	context.endFrame(1)
	context.renderAnim()
	try:
		os.rename(outpath+"0001.jpg",outpath+name_right)
	except OSError:
		os.remove(outpath+name_right)
		os.rename(outpath+"0001.jpg",outpath+name_right)	

	# on replace la camera a sa position initiale
	if cam_pos == 1:
		cam.LocX = cam.LocX - baseline
	else:
		cam.setLocation(loc[0],loc[1],loc[2])
	curScene.update()
	Blender.Redraw()
	Render.CloseRenderWindow()

if calc_disp == 1:

	# transformation world blender -> camera blender
	loc = cam.getLocation()
	C = [loc[0], loc[1], loc[2], 1]
	Rt_wb = STransformationMat(cam.getInverseMatrix())

	# transformation camera blender -> camera hartley
	Rt_cb = [[-1,0,0,0],[0,1,0,0],[0,0,-1,0],[0,0,0,1]]

	# transformation world blender -> camera hartley
	Rt = SMatMultMat(Rt_cb,Rt_wb)

	# calcul de la distance focale
	ratio = xres/yres
	fov = 2*atan(ratio * 16 / cam_data.getLens())

	foclen = xres/(2*tan(fov/2))

	# point principal
	x0 = (xres-1)/2.0
	y0 = (yres-1)/2.0

	print "***** equations des faces"
	tmats = getAllMats(Rt)
	print "***** equations des faces ok"

	# boucle sur les rayons
	LOIN = 1e10
	#fp = file(outpath+"p.dat",'w')
	#fq = file(outpath+"q.dat",'w')
	fdisp = file(outpath+"disp.dat",'w')
	for y in range(yres):
		print "***** ",floor(y*100/yres),"%"
		##y = 0
		yy = yres-y-1
		for x in range(xres):
			##x = 0
			xx = xres-x-1

			Zmin = LOIN
			nor = []

			### point 3d correspondant au point image courant, dans le repere camera
			X_ch = [xx-x0,yy-y0,foclen,1]
			C_ch = [0,0,0,1]

			# coordonnees de Plucker du rayon
			L = []
			L.append([X_ch[0],X_ch[1],X_ch[2]])
			L.append(SCrossVecs([X_ch[0],X_ch[1],X_ch[2]],[0,0,0]))

			# intersection de ce rayon avec toutes les faces
			for N in range(len(tmats)):
				PI = tmats[N][0]
				cotes = tmats[N][1]

				# y a t il intersection
				r = RayFaceIntersect(cotes,L)
				# si oui, on la calcule
				pt_int = []
				if r == 1:
					piD = [PI[0],PI[1],PI[2]]
					pt_int = SCrossVecs(L[1],piD)
					pt_int = [pt_int[0]-L[0][0]*PI[3],pt_int[1]-L[0][1]*PI[3],pt_int[2]-L[0][2]*PI[3]]
					pt_int.append(SDotVecs(L[0],piD))	
					SNormalise_homogene(pt_int)		

					if pt_int[2] < Zmin:
						Zmin = pt_int[2]
						pt_ch = [pt_int[0],pt_int[1],pt_int[2]]
			
			if Zmin < (LOIN-10):
				disp = -foclen * baseline / Zmin

				# ce point est-il dans l'image de droite ?
				xx2 = xx - disp
				if xx2 < eps or xx2 > xres:
					disp = 0
				else:
					# ce point est-t-il occulte ?
					# dans le repere camera gauche "ch" c le pt pt_z, a relier au centre optique droit en (-baseline,0,0)			
					C2 = [-baseline,0,0,1]
			
					# coordonnees de Plucker du rayon
					L = []
					L.append([pt_ch[0]-C2[0],pt_ch[1]-C2[1],pt_ch[2]-C2[2]])
					L.append(SCrossVecs([pt_ch[0],pt_ch[1],pt_ch[2]],[-baseline,0,0]))


					# intersection de ce rayon avec toutes les faces
					for N2 in range(len(tmats)):
						PI = tmats[N2][0]
						cotes = tmats[N2][1]
						# y a t il intersection
						r = RayFaceIntersect(cotes,L)
						# si oui, on la calcule
						if r == 1:
							piD = [PI[0],PI[1],PI[2]]
							pt_int = SCrossVecs(L[1],piD)
							pt_int = [pt_int[0]-L[0][0]*PI[3],pt_int[1]-L[0][1]*PI[3],pt_int[2]-L[0][2]*PI[3]]
							pt_int.append(SDotVecs(L[0],piD))	
							SNormalise_homogene(pt_int)		
							if pt_int[2] < (Zmin - eps):
								disp = 0
								break

			else:
				disp = "nan"

			fdisp.write(str(disp))
			fdisp.write(' ')

		gc.collect()
		fdisp.write('\n')
		fdisp.flush()
	fdisp.close()
	del C,Rt,curScene,cam_data,tmats
	gc.collect()

cam.setSize(cam_size[0],cam_size[1],cam_size[2])

print "****** fini "